from ._ey import CCA_EY, PLS_EY
from ._gha import CCA_GHA

from ._stochasticpls import PLSStochasticPower
from ._svd import CCA_SVD, PLS_SVD

__all__ = [
    "CCA_EY",
    "PLS_EY",
    "CCA_GHA",
    "CCA_SVD",
    "PLS_SVD",
    "PLSStochasticPower",
]
